/* global tinymce */
tinymce.PluginManager.add( 'linkEntitiesInline', function( editor ) {
	var linkButton,
		lastEntities = [],
		maxLastEntities = 6,
		lastStartOffset = 0,
		lastEndOffset = 0,
		DBPEDIA="http://dbpedia.org/sparql?default-graph-uri=http%3A%2F%2Fdbpedia.org&query=select+distinct+*+where+{<PUT-THE-IRI-HERE>+<http%3A%2F%2Fwww.w3.org%2F2000%2F01%2Frdf-schema%23comment>+%3Fc+.%0D%0AFILTER(langMatches(lang(%3Fc)%2C+'EN'))%0D%0A%0D%0A}+LIMIT+100&format=json&timeout=30000"
	
	editor.on( 'nodechange', function( event ) {
		setState( event );
	});

	editor.addCommand( 'WP_Entity_Unlink', function() {
		
		var parentNode = editor.selection.getNode();

		if ( parentNode.className == 'entityLink' ) {
			editor.dom.remove(parentNode, true);
		} else {
			var nodes = getNodesInRange(editor.selection.getRng());
			for ( var i=0; i < nodes.length; i++ ) {
				if ( nodes[i].className == 'entityLink' ) {
					editor.dom.remove(nodes[i], true);
				}
			}
			editor.selection.collapse();
		}

	});

	function setState( event ) {
		
		var node = event.element;
		
		var entityInSelection = false,
			sameSelection = false,
			nodes = getNodesInRange(editor.selection.getRng());

		for ( var i=0; i < nodes.length; i++ ) {
			if ( nodes[i].className == 'entityLink' ) {
				entityInSelection = true;
			}
		}

		if (editor.selection.getRng().startOffset == lastStartOffset && editor.selection.getRng().endOffset == lastEndOffset) {
			// TODO: FIX LOGIC
			if (jQuery('.suggestionContainer').length > 0) {
				sameSelection = true;
			}
		}

		lastStartOffset = editor.selection.getRng().startOffset;
		lastEndOffset = editor.selection.getRng().endOffset;

		if ( !editor.selection.isCollapsed() && node.className != 'entityLink' && !entityInSelection && !event.selectionChange && !sameSelection) {			
			
			// clean up
			closeUnlinkMenu();

			var queryText = editor.selection.getContent(),
				textPosition = getSelectedTextPosition(),
				wordCount = getWordCount(queryText);

			// prevent opening category menu when more than (refer_word_limit) words are selected
			if ( wordCount <= refer_word_limit ) {
				openCatMenu(queryText, textPosition);
			}

		} else if ( editor.selection.isCollapsed() && node.className != 'entityLink' && !entityInSelection ) {
			// clean up
			closeSuggestionContainer();
			closeUnlinkMenu();

		} else if ( node.className == 'entityLink' && !node.name && !event.selectionChange ) {
			// clean up
			closeSuggestionContainer();

			var textPosition = getSelectedTextPosition();
			openUnlinkMenu(textPosition);
		} else if ( entityInSelection && !node.name && !event.selectionChange ) {
			// Entity somewhere in selected text (dealt with by toolbar button)
		} else if ( event.selectionChange || sameSelection ) {
			// clean up
			closeSuggestionContainer();
			closeUnlinkMenu();
		}
	}

	function setButtonState( button, event ) {
		
		var node = event.element;
		
		var entityInSelection = false;
		var nodes = getNodesInRange(editor.selection.getRng());
		for ( var i=0; i < nodes.length; i++ ) {
			if ( nodes[i].className == 'entityLink' ) {
				entityInSelection = true;
			}
		}
				
		if (button.settings.cmd == 'WP_Entity_Unlink' && node.className != 'entityLink' && !entityInSelection) {
			button.disabled(true);
		} else {
			button.disabled( editor.selection.isCollapsed() && node.className != 'entityLink' && !entityInSelection );
		}
		button.active( button.settings.cmd == 'WP_Entity_Unlink' && (node.className == 'entityLink' || entityInSelection) && ! node.name );
	}

	function openCatMenu(query, position) {

		if ( window.referWorking ) return false;

		window.referWorking = true;

		var docBody = jQuery(editor.getContainer()).parents('body');

		// clean up
		closeSuggestionContainer();

		// add html container elements
		var suggestionContainer = jQuery('<div class="suggestionContainer working"></div>'),
			workingIndicator = jQuery('<div class="referWorkingIndicator"></div>'),
			circleMenuContainer = jQuery('<ul class="circleMenu"></ul>'),
			suggestionContent = jQuery('<div class="suggestionContent"></div>'),
			//suggestionModalBackBtn = jQuery('<div class="suggestionModalBack"></div>'),
			suggestionInput = jQuery('<input class="suggestionInput" value="'+ query +'"/>'),
			//suggestionModalCloseBtn = jQuery('<div class="suggestionModalClose"></div>'),
			suggestionDepiction = jQuery('<div class="suggestionDepiction"><div class="suggestionDepictionWrapper"></div></div>'),
			suggestionTabContainer = jQuery('<div class="suggestionTabContainer"></div>'),
			suggestionTabScrollContainer = jQuery('<div class="suggestionTabScrollContainer"></div>');

		suggestionTabContainer.append(suggestionTabScrollContainer);
		suggestionContent.append(suggestionInput, suggestionDepiction, suggestionTabContainer);
		suggestionContainer.append(workingIndicator, circleMenuContainer, suggestionContent);

		// init cat menu
		suggestionContainer.css({
			top: position.top + 'px',
			left: position.left + 'px'
		}).show();
		
		docBody.append(suggestionContainer);

		parseData(query, function() {

			// on success -> init event listeners
			jQuery('.suggestionContainer .circleMenu').circleMenu({
				item_diameter: 40,
				circle_radius: 70,
				angle: {
					start: 140,
					end: 310
				},
				trigger: 'none',
				select: function(evt, item) {
										
					if ( !item.hasClass('inactive') ) {
						jQuery('.suggestionContainer .suggestionContent').removeClass('modal contextModal').show();
						jQuery('.suggestionContainer').addClass('open');

						// init scrollable tab width
						jQuery('.suggestionContainer .suggestionTab').each(function() {
							var fullTabWidth = 0;

							jQuery(this).find('.suggestionTabControls').children('.suggestionTabButton').each(function() {
								fullTabWidth = fullTabWidth + jQuery(this).outerWidth(true);
							});

							jQuery(this).find('.suggestionTabControls').width(fullTabWidth);
						});

						selectMetaCategory(item.data('category'));

					}
					

					evt.stopPropagation();
					evt.preventDefault();
					return false;
					
				},
				speed: 120,
				step_out: 0
			});

			var keystrokeTimeout;

			if (lastEntities.length > 0) {
				jQuery('.circleMenuItem.recentlySelected').removeClass('inactive').attr('title', 'Recently Selected');
			} else {
				jQuery('.circleMenuItem.recentlySelected').addClass('inactive').attr('title', 'Recently Selected (empty)');
			}

			jQuery('.suggestionInput').keyup(function(evt) {
				
				if ( jQuery(this).val().length > 2 && evt.keyCode != 37 && evt.keyCode != 38 && evt.keyCode != 39 && evt.keyCode != 40 && evt.keyCode != 13 ) {
					clearTimeout(keystrokeTimeout);
					keystrokeTimeout = setTimeout(function() {
						updateSuggestions( jQuery('.suggestionInput').val() );
					}, 500);
				} else {
					//
				}

			})

			jQuery('.suggestionTabButton').click(function(evt) {
				selectClassTab( jQuery(this).index() );
			});

			jQuery('.suggestionContainer .circleMenu').circleMenu('open');
			jQuery('.circleMenuItem').not('.root, .recentlySelected').donutchart('animate');

			// init click listener which inserts the entity
			jQuery('.suggestionContainer .suggestedEntity').click(function(evt) {

				var parentNode = editor.selection.getNode(),
					uriValue,
					labelValue,
					dataLabel = jQuery(this).data('label'),
					dataUri = jQuery(this).data('uri'),
					dataCategory = jQuery(this).data('category'),
					dataThumbnail = jQuery(this).data('thumbnail');


				if ( parentNode.className == 'entityLink' && parentNode.getAttribute('data-uri').length > 0 ) {
					uriValue = parentNode.getAttribute('data-uri');
					labelValue = parentNode.getAttribute('data-label');
				} else {
					uriValue = '';
					labelValue = editor.selection.getContent();
				}

				if ( parentNode.className == 'entityLink' && dataUri === '' ) {
					editor.dom.remove(parentNode, true);
				} else if ( parentNode.className == 'entityLink' ) {
					parentNode.setAttribute('data-uri', dataUri);
				} else {
					var selected_text = editor.selection.getContent();
					var type = 'http://dbpedia.org/ontology/'+ dataCategory ;
	                if (dataCategory=='http://www.w3.org/2002/07/owl#Thing'){
	                        type = 'http://www.w3.org/2002/07/owl#Thing';
	                }
	                
	                var thumbnailString = '';
					if ( dataThumbnail ) {
	                	thumbnailString = 'data-thumbnail="'+ dataThumbnail +'"';
	                }

	                var return_text = '<span class="entityLink" property="http://www.w3.org/2004/02/skos/core#subject" data-createdby="user" typeof="' + type + '" resource="'+ dataUri +'" data-uri="'+ dataUri +'" data-category="'+ dataCategory +'" '+ thumbnailString +' data-label="'+ dataLabel +'">' + selected_text + '</span>';

			        editor.execCommand('mceInsertContent', 0, return_text );
			    }

			    // fill into lastElements array (via tmp Array)
				// (1) put it into the new array  
				// (2) fill with other element entries, skipping the identical one (if existing) 					
				var tmp = [],
					currentEntity = {
						'label': dataLabel,
						'iri': dataUri,
						'categoryIri': dataCategory,
						'thumbnailUri': dataThumbnail
					};
				tmp.push(currentEntity);
				for (var k = 0 ; k < Math.min(lastEntities.length, maxLastEntities) ; k++ ){
					if (currentEntity.iri != lastEntities[k].iri){
						tmp.push(lastEntities[k]);
					}
				}
				lastEntities = tmp;
				

			}).hover(function(evt) {
				
				jQuery(this).parent().children('.suggestedEntity').removeClass('active');
				jQuery(this).addClass('active');

				var dataThumbnail = jQuery(this).data('thumbnail'),
					dataCategory = jQuery(this).data('category');
				setThumbnailPreview(dataThumbnail, dataCategory);

			}, function(evt) {
				
				jQuery(this).removeClass('active');
				setThumbnailPreview(false);
			});

			jQuery(document).on('keydown', function(evt) {

				switch (evt.keyCode) {
					case 38:
						interfaceUp();
						evt.stopPropagation();
						evt.preventDefault();
						break;
			        case 40:
			        	interfaceDown();
			        	evt.stopPropagation();
						evt.preventDefault();
			        	break;
			        case 37:
			        	interfaceLeft();
			        	break;
			        case 39:
			        	interfaceRight();
			        	break;
			        default:
			        	// default
				}

			});

			jQuery(document).on('keyup', function(evt) {

				switch (evt.keyCode) {
					case 13:
						
						var selectedEntity = getSelectedEntity().entity;

						if ( selectedEntity && evt.target.className != 'suggestionInput' ) {
							selectedEntity.click();
							evt.stopPropagation();
							evt.preventDefault();
						}
						
						break;
			        default:
			        	// default
				}

			});

			window.referWorking = false;
			jQuery('.suggestionContainer').removeClass('working');

		}, function() {

			// on error
			window.referWorking = false;

		});

	}

	function updateSuggestions(query) {

		if (window.request) {
			window.request.abort();
		}
		
		window.referWorking = true;
		
		var docBody = jQuery(editor.getContainer()).parents('body');
		
		// clean up
		jQuery('.suggestionContainer').addClass('working');
		jQuery('.suggestionContainer .circleMenuItem').not('.recentlySelected').removeClass('inactive').removeAttr('title');
		jQuery('.suggestionTabScrollContainer').empty();

		parseData(query, function() {

			jQuery('.suggestionTab[data-category="'+ jQuery('.circleMenuItem.active').data('category') +'"]').addClass('active');
			
			// init scrollable tab width
			jQuery('.suggestionContainer .suggestionTab').each(function() {
				var fullTabWidth = 0;

				jQuery(this).find('.suggestionTabControls').children('.suggestionTabButton').each(function() {
					fullTabWidth = fullTabWidth + jQuery(this).outerWidth(true);
				});

				jQuery(this).find('.suggestionTabControls').width(fullTabWidth);
			});

			selectClassTab(0);


			// on success -> init event listeners
			jQuery('.suggestionTabButton').click(function(evt) {
				selectClassTab( jQuery(this).index() );
			});

			jQuery('.circleMenuItem').not('.root, .recentlySelected').donutchart('animate');

			// init click listener which inserts the entity
			jQuery('.suggestionContainer .suggestedEntity').click(function(evt) {

				var parentNode = editor.selection.getNode(),
					uriValue,
					labelValue,
					dataLabel = jQuery(this).data('label'),
					dataUri = jQuery(this).data('uri'),
					dataCategory = jQuery(this).data('category'),
					dataThumbnail = jQuery(this).data('thumbnail');


				if ( parentNode.className == 'entityLink' && parentNode.getAttribute('data-uri').length > 0 ) {
					uriValue = parentNode.getAttribute('data-uri');
					labelValue = parentNode.getAttribute('data-label');
				} else {
					uriValue = '';
					labelValue = editor.selection.getContent();
				}

				if ( parentNode.className == 'entityLink' && dataUri === '' ) {
					editor.dom.remove(parentNode, true);
				} else if ( parentNode.className == 'entityLink' ) {
					parentNode.setAttribute('data-uri', dataUri);
				} else {
					var selected_text = editor.selection.getContent();
					var type = 'http://dbpedia.org/ontology/'+ dataCategory ;
	                if (dataCategory=='http://www.w3.org/2002/07/owl#Thing'){
	                        type = 'http://www.w3.org/2002/07/owl#Thing';
	                }
	                
	                var thumbnailString = '';
					if ( dataThumbnail ) {
	                	thumbnailString = 'data-thumbnail="'+ dataThumbnail +'"';
	                }

	                var return_text = '<span class="entityLink" property="http://www.w3.org/2004/02/skos/core#subject" data-createdby="user" typeof="' + type + '" resource="'+ dataUri +'" data-uri="'+ dataUri +'" data-category="'+ dataCategory +'" '+ thumbnailString +' data-label="'+ dataLabel +'">' + selected_text + '</span>';

			        editor.execCommand('mceInsertContent', 0, return_text );
			    }

			}).hover(function(evt) {
				
				jQuery(this).parent().children('.suggestedEntity').removeClass('active');
				jQuery(this).addClass('active');

				var dataThumbnail = jQuery(this).data('thumbnail'),
					dataCategory = jQuery(this).data('category');
				setThumbnailPreview(dataThumbnail, dataCategory);

			}, function(evt) {
				
				jQuery(this).removeClass('active');
				setThumbnailPreview(false);
			});

			window.referWorking = false;
			jQuery('.suggestionContainer').removeClass('working');

		}, function() {

			// on error
			window.referWorking = false;

		}, true);

	}

	function closeSuggestionContainer() {
		if (window.request) {
			window.request.abort();
		}

		jQuery(editor.getContainer()).parents('body').find('.suggestionContainer').remove();

		jQuery(document).off('keydown');
		jQuery(document).off('keyup');
	}

	function selectMetaCategory(catName) {
		
		deselectEntities();

		jQuery('.suggestionContainer .circleMenu li').removeClass('active');
		jQuery('.suggestionContainer .circleMenu li[data-category="'+ catName +'"]').addClass('active');

		var catTabElem = jQuery('.suggestionTab[data-category="'+ catName +'"]'),
			catIndex = catTabElem.index(),
			catOffset = catTabElem.height() * catIndex;

		jQuery('.suggestionTab').removeClass('active');
		catTabElem.addClass('active');
		
		jQuery('.suggestionTabScrollContainer').css('margin-top', - catOffset + 'px');

		if ( catTabElem.find('.suggestionTabButton.active').length == 0 ) {
			selectClassTab(0);
		}
	}

	function selectMetaCategoryPrev() {
		setThumbnailPreview(false);

		if ( jQuery('.suggestionTab.active').length == 0 ) { 
			return;
		}

		var activeCat = jQuery('.suggestionTab.active'),
			prevCat = activeCat.prev('.suggestionTab');

		if ( prevCat.length == 0 ) {
			return;
		}

		selectMetaCategory(prevCat.data('category'));
	}

	function selectMetaCategoryNext() {
		setThumbnailPreview(false);

		if ( jQuery('.suggestionTab.active').length == 0 ) {
			return;
		}

		var activeCat = jQuery('.suggestionTab.active'),
			nextCat = activeCat.next('.suggestionTab');

		if ( nextCat.length == 0 ) {
			return;
		}

		selectMetaCategory(nextCat.data('category'));
	}

	function selectEntity(index) {
		var entityList = getSelectedEntity().activeClassTab,
			entityElem = entityList.children('.suggestedEntity').eq(index);

		// clean up
		entityList.children('.suggestedEntity').removeClass('active');

		var newScrollPosition = entityElem.offset().top - entityList.offset().top + entityList.scrollTop() - entityList.height()/2 + entityElem.height()/2;
		
		entityList.animate({
		    scrollTop: newScrollPosition
		}, 100);

		entityElem.addClass('active');
		setThumbnailPreview(entityElem.data('thumbnail'), entityElem.data('category'));
	}

	function selectEntityPrev() {
		var activeEntity = getSelectedEntity(),
			activeEntityElem;

		if ( !activeEntity.entity ) { 
			selectEntity(0);
			return;
		}

		activeEntityElem = activeEntity.entity;

		var prevEntity = activeEntityElem.prev('.suggestedEntity');
		if ( prevEntity.length == 0 ) {
			return;
		}

		selectEntity(prevEntity.index());
	}

	function selectEntityNext() {
		var activeEntity = getSelectedEntity(),
			activeEntityElem;

		if ( !activeEntity.entity ) { 
			selectEntity(0);
			return;
		}

		activeEntityElem = activeEntity.entity;

		var nextEntity = activeEntityElem.next('.suggestedEntity');
		if ( nextEntity.length == 0 ) {
			return;
		}

		selectEntity(nextEntity.index());
	}

	function getSelectedEntity() {
		var activeMetaCatElem = jQuery('.suggestionTab.active'),
			activeClassTab = activeMetaCatElem.find('.suggestionTabEntityList.active');
			
		if ( activeMetaCatElem.length == 0 || activeClassTab.length == 0 ) { 
			return {
				entity: null,
				activeClassTab: null,
				activeMetaCatElem: null
			}
		}

		var activeEntity = activeClassTab.children('.suggestedEntity.active');
		if ( activeEntity.length == 0 ) {
			return {
				entity: null,
				activeClassTab: activeClassTab,
				activeMetaCatElem: activeMetaCatElem
			}
		} else {
			return {
				entity: activeEntity,
				activeClassTab: activeClassTab,
				activeMetaCatElem: activeMetaCatElem
			};
		}
	}

	function deselectEntities() {
		jQuery('.suggestionTabContainer .suggestedEntity.active').removeClass('active');
		setThumbnailPreview(false);
	}

	function selectClassTab(index) {

		var activeMetaCatElem = jQuery('.suggestionTab.active');
		if ( activeMetaCatElem.length == 0 ) {
			deselectEntities();
			return;
		}

		deselectEntities();

		// activate tab
		var activeClassTabButton = activeMetaCatElem.find('.suggestionTabButton').eq(index);

		activeMetaCatElem.find('.suggestionTabButton').removeClass('active');
		
		if (activeClassTabButton.length > 0) {
			activeClassTabButton.addClass('active');

			var currentTabWidth = activeClassTabButton.width(),
				tabControlsWrapper = activeMetaCatElem.find('.suggestionTabControlsWrapper'),
				tabControls = tabControlsWrapper.children('.suggestionTabControls'),
				scrollButtonLeft = tabControlsWrapper.children('.suggestionTabScrollButton.left'),
				scrollButtonRight = tabControlsWrapper.children('.suggestionTabScrollButton.right'),
				newControlsPosition = - activeClassTabButton.position().left + tabControlsWrapper.width()/2 - currentTabWidth/2;
			
			tabControlsWrapper.children('.suggestionTabScrollButton').addClass('active');

			if (index == 0) {
				scrollButtonLeft.removeClass('active');
			} else if (index == tabControls.children('.suggestionTabButton').length - 1) {
				scrollButtonRight.removeClass('active');
			}

			if ( newControlsPosition > 0 ) {
				// max left
				tabControls.css('left', 0 + 'px');
			} else if ( newControlsPosition < - ( tabControls.width() - tabControlsWrapper.width() ) && tabControls.width() > tabControlsWrapper.width() ) {
				// max right
				tabControls.css('left', - ( tabControls.width() - tabControlsWrapper.width() ) + 'px');
			} else if ( tabControls.width() > tabControlsWrapper.width() ) {
				tabControls.css('left', newControlsPosition + 'px');
			}

		}

		
		// activate entity list
		var classTabElem = activeMetaCatElem.find('.suggestionTabEntityList').eq(index),
			classTabOffset = classTabElem.width() * index;

		activeMetaCatElem.find('.suggestionTabEntityList').removeClass('active');
		classTabElem.addClass('active');
		
		activeMetaCatElem.find('.suggestionTabEntityListScrollContainer').css('margin-left', - classTabOffset + 'px');
	}

	function selectClassTabPrev() {
		deselectEntities();

		var activeMetaCatElem = jQuery('.suggestionTab.active');
		if ( activeMetaCatElem.length == 0 ) {
			return;
		}

		if ( activeMetaCatElem.find('.suggestionTabButton.active').length == 0 ) { 
			return;
		}

		var activeClassTab = activeMetaCatElem.find('.suggestionTabButton.active'),
			prevClassTab = activeClassTab.prev('.suggestionTabButton');

		if ( prevClassTab.length == 0 ) {
			return;
		}

		selectClassTab(prevClassTab.index());
	}

	function selectClassTabNext() {
		deselectEntities();

		var activeMetaCatElem = jQuery('.suggestionTab.active');
		if ( activeMetaCatElem.length == 0 ) {
			return;
		}

		if ( activeMetaCatElem.find('.suggestionTabButton.active').length == 0 ) { 
			return;
		}

		var activeClassTab = activeMetaCatElem.find('.suggestionTabButton.active'),
			nextClassTab = activeClassTab.next('.suggestionTabButton');

		if ( nextClassTab.length == 0 ) {
			return;
		}

		selectClassTab(nextClassTab.index());
	}

	function cleanupData(apiData) {

		var cleanData = apiData,
			similarCategoryMapping = {
				'Agent': 'Person',
				'FilmFestival': 'Festival',
				'MusicFestival': 'Festival',
				'Senator': 'Politician',
				'OfficeHolder': 'Politician',
				'MemberOfParliament': 'Politician',
				'Congressman': 'Politician',
				'ComicsCharacter': 'FictionalCharacter',
				'NationalCollegiateAthleticAssociationAthlete': 'Athlete',
				'BasketballPlayer': 'Athlete',
				'BaseballPlayer': 'Athlete',
				'RugbyPlayer': 'Athlete',
				'TennisPlayer': 'Athlete',
				'SoccerPlayer': 'Athlete',
				'AmericanFootballPlayer': 'Athlete',
				'GridironFootballPlayer': 'Athlete',
				'Boxer': 'Athlete',
				'Cyclist': 'Athlete',
				'IceHockeyPlayer': 'Athlete',
				'GolfPlayer': 'Athlete',
				'AustralianRulesFootballPlayer': 'Athlete',
				'GaelicGamesPlayer': 'Athlete',
				'Cricketer': 'Athlete',
				'BadmintonPlayer': 'Athlete',
				'SnookerChamp': 'SnookerPlayer',
				'Theatre': 'ArchitecturalStructure',
				'Museum': 'ArchitecturalStructure',
				'Building': 'ArchitecturalStructure',
				'Stadium': 'ArchitecturalStructure',
				'StadiumOrArena': 'ArchitecturalStructure',
				'ShoppingMall': 'ArchitecturalStructure',
				'ShoppingCenter': 'ArchitecturalStructure',
				'Settlement': 'Place',
				'PopulatedPlace': 'Place',
				'ProtectedArea': 'Place',
				'AdministrativeRegion': 'AdministrativeArea',
				'Film': 'Movie',
				'TVEpisode': 'TelevisionEpisode',
				'Work': 'CreativeWork',
				'PeriodicalLiterature': 'WrittenWork',
				'Book': 'WrittenWork',
				'Magazine': 'WrittenWork',
				'Mountain': 'NaturalPlace',
				'LunarCrater': 'NaturalPlace',
				'Bridge': 'Infrastructure',
				'RouteOfTransportation': 'Infrastructure',
				'Road': 'Infrastructure',
				'RoadTunnel': 'Infrastructure',
				'Tunnel': 'Infrastructure',
				'RailwayLine': 'Infrastructure',
				'PublicTransitSystem': 'Infrastructure',
				'SkiArea': 'SkiResort',
				'HistoricBuilding': 'HistoricPlace',
				'LandmarksOrHistoricalBuildings': 'HistoricPlace',
				'Ship': 'MeanOfTransportation',
				'Lake': 'BodyOfWater',
				'LakeBodyOfWater': 'BodyOfWater',
				'River': 'BodyOfWater',
				'RiverBodyOfWater': 'BodyOfWater',
				'Stream': 'BodyOfWater',
				'Canal': 'BodyOfWater',
				'Organization': 'Organisation',
				'Company': 'Organisation',
				'Band': 'MusicGroup',
				'MusicalArtist': 'MusicGroup',
				'MusicAlbum': 'MusicalWork',
				'Website': 'WebPage',
				'TelevisionStation': 'Broadcaster',
				'WomensTennisAssociationTournament': 'SportsEvent',
				'WrestlingEvent': 'SportsEvent',
				'MixedMartialArtsEvent': 'SportsEvent',
				'Olympics': 'SportsEvent',
				'BritishRoyalty': 'Royalty',
				'AdultActor': 'Actor',
				'Mollusca': 'Animal',
				'Insect': 'Animal',
				'Eukaryote': 'Species',
				'Saint': 'Cleric',
				'ChristianBishop': 'Cleric',
				'University': 'EducationalInstitution',
				'CollegeOrUniversity': 'EducationalInstitution',
				'EducationalOrganization': 'EducationalInstitution',
				'GovernmentAgency': 'GovernmentOrganization',
				'Muscle': 'AnatomicalStructure'
			},
			entitiesToRemove = [];

		// iterate entities
		for (var e=0; e < cleanData.entities.length; e++) {
					
			var metaCat = cleanData.entities[e].categoryIri,
				catLabel = cleanData.entities[e].categoryLabel.split(','),
				cleanCatLabels = [];
			
			for (var cl=0; cl < catLabel.length; cl++) {
				// remove prefixes
				catLabel[cl] = catLabel[cl].replace('http://schema.org/','')
										.replace('http://xmlns.com/foaf/0.1/','')
										.replace('http://www.w3.org/2002/07/owl#','')
										.replace('http://purl.org/ontology/bibo/', '');
				// merge similar categories 
				if (similarCategoryMapping[catLabel[cl]]) {
					catLabel[cl] = similarCategoryMapping[catLabel[cl]];
				}

				// remove duplicates
				if ( jQuery.inArray(catLabel[cl], cleanCatLabels) == -1 ) {
					cleanCatLabels.push(catLabel[cl]);
				}
			}

			for (var cl = cleanCatLabels.length -1; cl >= 0; cl--) {
				// remove categories with the same label as the metaCat
				if (metaCat != 'http://www.w3.org/2002/07/owl#Thing' && cleanCatLabels[cl] == 'Thing') {
					if (cleanCatLabels.length > 1) {
						cleanCatLabels.splice(cl, 1);	
					} else {
						// if owl#Thing is the only category in a metaCat other than owl#Thing, remove the entire entity
						entitiesToRemove.push(e);
					}	
				} else if (cleanCatLabels.length > 1 && cleanCatLabels[cl] == metaCat) {
					// if there are additional categories, remove the metaCat (Person, Place, Event)
					cleanCatLabels.splice(cl, 1);
				}
			}

			cleanData.entities[e].categoryLabel = cleanCatLabels.join(',');
			//console.log(cleanData.entities[e].categoryLabel);
		}

		// clean up entities (in reverse order to not mess up the entity array within the loop)
		for (var r = entitiesToRemove.length -1; r >= 0; r--) {
			cleanData.entities.splice(entitiesToRemove[r], 1);
		}

		return cleanData;
	}

	function parseData(query, successCallback, errorCallback, reInit) {

		var cleanQuery = query.split('.').join('');

		window.request = jQuery.ajax({
			url: "http://apps.yovisto.com/refer-rest/services/suggest/x?input="+ cleanQuery,
			crossDomain: true,
			beforeSubmit: function() {
				
			}
		})
		.done(function(data) {
			
			if (data.error) {
				errorCallback();
			}

			data = cleanupData(data);
			
			var entities = data.entities,
				dynamicMetaCats = [];

			// don't fetch Meta Categories automatically (static list)
			var metaCats = ["Person", "Place", "Event", "http://www.w3.org/2002/07/owl#Thing"],
				categoryCnt = {
					"Person": 0,
					"Place": 0,
					"Event": 0,
					"http://www.w3.org/2002/07/owl#Thing": 0
				},
				totalEntities = 0;

			for (var c=0; c < entities.length; c++) {
				if ( jQuery.inArray(entities[c].categoryIri, dynamicMetaCats) == -1 ) {
					dynamicMetaCats.push(entities[c].categoryIri);
				}

				categoryCnt[entities[c].categoryIri]++;
				totalEntities++;
			}

			// Parse MetaCategories
			for (var m = 0; m < metaCats.length; m++) {				

				var resultPercent = (categoryCnt[metaCats[m]] / totalEntities) * 100,
					menuItem;

				if (reInit) {
					menuItem = jQuery('.circleMenuItem[data-category="'+ metaCats[m] +'"]');
					menuItem.attr('data-percent', resultPercent);
				} else {
					menuItem = jQuery('<li class="circleMenuItem" data-percent="'+ resultPercent +'" title="'+ metaCats[m] +'" data-category="'+ metaCats[m] +'"></li>');
				}
				
				if ( jQuery.inArray(metaCats[m], dynamicMetaCats) == -1 ) {
					// no results for this Meta Category
					menuItem.addClass('inactive').attr({
						'data-percent': 0,
						'title': 'No Results in this Category'
					});
				}

				if (!reInit) {
					jQuery('.circleMenu').prepend(menuItem);
					menuItem.donutchart({
						'bgColor' : 'transparent',
					    'fgColor' : 'rgba(255,255,255,.3)',
					    'size' : 40
					});
				}

				/*
				var startMenuItem = jQuery('<div class="startCircleMenuItem" data-percent="'+ resultPercent +'" data-tooltip="'+ metaCats[m] +'" data-category="'+ metaCats[m] +'"></div>');
				jQuery('.suggestionStartContainer .suggestionStartCategories').append(startMenuItem);

				startMenuItem.donutchart({
					'bgColor' : 'transparent',
				    'fgColor' : '#efefef',
				    'size' : 40
				});
				*/

				var suggestionTab = jQuery('<div class="suggestionTab" data-category="'+ metaCats[m] +'"></div>'),
					suggestionTabControlsWrapper = jQuery('<div class="suggestionTabControlsWrapper"></div>'),
					suggestionTabControls = jQuery('<div class="suggestionTabControls"></div>'),
					suggestionTabScrollButtonLeft = jQuery('<div class="suggestionTabScrollButton left" title="Tip: Use Arrow Keys to Navigate" data-category="'+ metaCats[m] +'"></div>'),
					suggestionTabScrollButtonRight = jQuery('<div class="suggestionTabScrollButton right" title="Tip: Use Arrow Keys to Navigate" data-category="'+ metaCats[m] +'"></div>'),
					suggestionTabEntityListContainer = jQuery('<div class="suggestionTabEntityListContainer"></div>'),
					suggestionTabEntityListScrollContainer = jQuery('<div class="suggestionTabEntityListScrollContainer"></div>');

				suggestionTabScrollButtonLeft.click(selectClassTabPrev);
				suggestionTabScrollButtonRight.click(selectClassTabNext);

				var dynamicClasses = [];

				// assemble and cleanup dynamic classes
				for (var e=0; e < entities.length; e++) {
					
					if (entities[e].categoryIri == metaCats[m]) {
						var catLabel = entities[e].categoryLabel.split(',');
						for (cl=0; cl < catLabel.length; cl++) {
							// add class
							if ( jQuery.inArray(catLabel[cl], dynamicClasses) == -1 ) {
								dynamicClasses.push(catLabel[cl]);
							}
						}
					}
				}

				// iterate through dynamic classes
				for (var c = 0; c < dynamicClasses.length; c++) {
					
					var suggestionTabButton = jQuery('<div class="suggestionTabButton" data-class="'+ dynamicClasses[c] +'">'+ dynamicClasses[c] +'</div>'),
						suggestionTabEntityList = jQuery('<ul class="suggestionTabEntityList" data-class="'+ dynamicClasses[c] +'"></ul>');

					suggestionTabControls.append(suggestionTabButton);
					
					// iterate though entities to find entity classes
					for (var e=0; e < entities.length; e++) {
						var catLabel = entities[e].categoryLabel.split(',');

						// check if current dynamic class appears in entity class array and append entity element
						if ( jQuery.inArray(dynamicClasses[c], catLabel) != -1 ) {
							var entityElement = jQuery('<li class="suggestedEntity" title="'+ entities[e].label +'" data-category="'+ metaCats[m] +'" data-uri="'+ entities[e].iri +'" data-label="'+ entities[e].label +'"><span class="suggestedEntityLabel">'+ entities[e].label +'</span><span class="suggestedEntityUri">'+ entities[e].iri +'</span></li>');
							if ( entities[e].thumbnailUri ) {
								entityElement.attr('data-thumbnail', entities[e].thumbnailUri);
							} 
							suggestionTabEntityList.append(entityElement);

						}
					}

					suggestionTabEntityListScrollContainer.append(suggestionTabEntityList);

				}
				

				suggestionTabEntityListContainer.append(suggestionTabEntityListScrollContainer);

				suggestionTabControlsWrapper.append(suggestionTabControls);
				suggestionTabControlsWrapper.append(suggestionTabScrollButtonLeft, suggestionTabScrollButtonRight);
				suggestionTab.append(suggestionTabControlsWrapper);
				suggestionTab.append(suggestionTabEntityListContainer);

				jQuery('.suggestionTabScrollContainer').append(suggestionTab);

			}

			// add recently selected elements

			var recentlySelectedContainer = jQuery('<div class="suggestionTab recentlySelectedContainer" data-category="recent"><span>Recently Selected:</span></div>'),
				recentlySelectedEntityListContainer = jQuery('<div class="suggestionTabEntityListContainer"></div>'),
				recentlySelectedEntityListScrollContainer = jQuery('<div class="suggestionTabEntityListScrollContainer"></div>'),
				recentlySelectedContent = jQuery('<ul class="suggestionTabEntityList recentlySelectedContent active"></ul>');
				
			// fill recently selected container
			for (var i=0; i < lastEntities.length; i++){
				var currentEntity = lastEntities[i],
					entityElement = jQuery('<li class="suggestedEntity" title="'+ currentEntity.label +'" data-category="'+ currentEntity.categoryIri +'" data-uri="'+ currentEntity.iri +'" data-label="'+ currentEntity.label +'"><span class="suggestedEntityLabel">'+ currentEntity.label +'</span><span class="suggestedEntityUri">'+ currentEntity.iri +'</span></li>');
				
				if ( currentEntity.thumbnailUri ) {
					entityElement.attr('data-thumbnail', currentEntity.thumbnailUri);
				}
				entityElement.appendTo(recentlySelectedContent);
			}

			recentlySelectedContent.appendTo(recentlySelectedEntityListScrollContainer);
			recentlySelectedEntityListScrollContainer.appendTo(recentlySelectedEntityListContainer);
			recentlySelectedEntityListContainer.appendTo(recentlySelectedContainer);

			jQuery('.suggestionTabScrollContainer').append(recentlySelectedContainer);


			if (!reInit) {
				jQuery('.circleMenu').prepend('<li class="circleMenuItem recentlySelected" data-category="recent"></li>');
				jQuery('.circleMenu').prepend('<li class="circleMenuItem root"></li>');
			}

			successCallback();

		})
		.fail(function() {
			errorCallback();
		})
		.always(function() {
			//
		});	

	}

	function openUnlinkMenu(position) {
		// clean up
		closeUnlinkMenu();

		var docBody = jQuery(editor.getContainer()).parents('body');

		// add html container elements
		var unlinkMenuContainer = jQuery('<div class="referUnlinkMenuContainer"></div>'),
			unlinkToolbar = jQuery('<div role="group" class="mce-toolbar-grp mce-inline-toolbar-grp mce-container mce-panel mce-arrow-down mce-arrow-left" style="position: relative;"></div>'),
			unlinkBody = jQuery('<div class="mce-toolbar mce-container-body mce-stack-layout"></div>'),
			unlinkButtonGroupWrapper = jQuery('<div role="toolbar" class="mce-container mce-stack-layout-item mce-first mce-last mce-btn-group"></div>'),
			//unlinkButtonEdit = jQuery('<div tabindex="-1" class="mce-widget mce-btn" role="button" aria-label="Edit "><button tabindex="-1" type="button" role="presentation"><i class="mce-ico mce-i-dashicon dashicons-edit"></i></button></div>'),
			unlinkButtonRemove = jQuery('<div tabindex="-1" class="mce-widget mce-btn mce-last" role="button" aria-label="Remove"><button tabindex="-1" type="button" role="presentation"><i class="mce-ico mce-i-dashicon dashicons-no"></i></button></div>'),
			nodeUri = jQuery(editor.selection.getNode()).attr('resource'),
			nodeCategory = jQuery(editor.selection.getNode()).data('category'),
			unlinkButtonUri = jQuery('<a href="'+ nodeUri +'" target="_blank" style="display: inline-block; white-space: nowrap; color: inherit; font-size: 12px; line-height: 12px; padding: 8px 12px 8px 4px;">'+ nodeUri +'</a>');

		unlinkButtonGroupWrapper.append(unlinkButtonRemove, unlinkButtonUri);
		unlinkBody.append(unlinkButtonGroupWrapper);
		unlinkToolbar.append(unlinkBody);
		unlinkMenuContainer.append(unlinkToolbar);

		// init unlink menu
		unlinkButtonRemove.click(function(evt) {
			editor.execCommand('WP_Entity_Unlink');
		});

		if ( nodeCategory ) {
			unlinkMenuContainer.attr('data-category', nodeCategory);
		}

		unlinkMenuContainer.css({
			top: position.top + 'px',
			left: position.left + 'px'
		}).show();
		
		docBody.append(unlinkMenuContainer);
	}

	function closeUnlinkMenu() {
		jQuery(editor.getContainer()).parents('body').find('.referUnlinkMenuContainer').remove();
	}

	function setThumbnailPreview(source, category) {
		if ( source === false ) {
			jQuery('.suggestionContainer .suggestionDepiction').removeClass('active');
			jQuery('.suggestionContainer .suggestionDepictionImg').remove();
		} else {
			jQuery('.suggestionContainer .suggestionDepictionImg').remove();
			jQuery('.suggestionContainer .suggestionDepiction').addClass('active');
			if (source) {
				jQuery('.suggestionContainer .suggestionDepiction .suggestionDepictionWrapper').attr('data-category', category).append('<img class="suggestionDepictionImg" src="'+ source +'"/>');
			} else {
				jQuery('.suggestionContainer .suggestionDepiction .suggestionDepictionWrapper').attr('data-category', category).append('<span class="suggestionDepictionImg"></span>');
			}
		}
	}

	/*******************
	* Helper Methods
	*******************/

	function getWordCount(textString) {
		textString = textString.replace(/(^\s*)|(\s*$)/gi,"");//exclude  start and end white-space
	    textString = textString.replace(/[ ]{2,}/gi," ");//2 or more space to 1
	    textString = textString.replace(/\n /,"\n"); // exclude newline with a start spacing
		return textString.split(' ').length;
	}

	function getSelectedTextPosition() {
		var tinymcePosition = jQuery(editor.getContainer()).position(),
			toolbarPosition = jQuery(editor.getContainer()).find(".mce-toolbar").first(),
			nodePosition = jQuery(editor.selection.getNode()).position(),
			textareaTop = 0,
			textareaLeft = 0;

			if (editor.selection.getRng().getClientRects().length > 0) {
			    textareaTop = parseInt(jQuery(editor.selection.getNode()).css("font-size")) * 0.4 + editor.selection.getRng().getClientRects()[0].top;
			    textareaLeft = editor.selection.getRng().getClientRects()[0].left;
			} else {
			    textareaTop = parseInt(jQuery(editor.selection.getNode()).css("font-size")) * 0.4 + nodePosition.top;
			    textareaLeft = nodePosition.left;
			}

			var position = jQuery(editor.getContainer()).offset();
			var caretPosition = {
			    top:  tinymcePosition.top + toolbarPosition.innerHeight() + textareaTop + position.top,
			    left: tinymcePosition.left + textareaLeft + position.left
			}

		return caretPosition;
	}

	function getNextNode(node) {
	    if (node.firstChild)
	        return node.firstChild;
	    while (node) {
	        if (node.nextSibling)
	            return node.nextSibling;
	        node = node.parentNode;
	    }
	}

	function getNodesInRange(range) {
	    var start = range.startContainer;
	    var end = range.endContainer;
	    var commonAncestor = range.commonAncestorContainer;
	    var nodes = [];
	    var node;

	    // walk parent nodes from start to common ancestor
	    for (node = start.parentNode; node; node = node.parentNode) {
	        nodes.push(node);
	        if (node == commonAncestor)
	            break;
	    }
	    nodes.reverse();

	    // walk children and siblings from start until end is found
	    for (node = start; node; node = getNextNode(node)) {
	        nodes.push(node);
	        if (node == end)
	            break;
	    }

	    return nodes;
	}

	/*******************
	* Key Control Methods
	*******************/

	function interfaceUp() {
		selectEntityPrev();
	}

	function interfaceDown() {
		selectEntityNext();
	}

	function interfaceLeft() {
		selectClassTabPrev();
	}

	function interfaceRight() {
		selectClassTabNext();
	}

	editor.addButton( 'unlinkEntity', {
		icon: 'unlinkEntity',
		tooltip: 'Remove entity',
		cmd: 'WP_Entity_Unlink',
		onPostRender: function() {
			var unlinkButton = this;

			editor.on( 'nodechange', function( event ) {
				setButtonState(unlinkButton, event);
			});
		}
	});

});
