/* global tinymce */
tinymce.PluginManager.add( 'linkEntities', function( editor ) {
	var linkButton;
	var lastEntities = [];
	var maxLastEntities = 5;
	
	var DBPEDIA="http://dbpedia.org/sparql?default-graph-uri=http%3A%2F%2Fdbpedia.org&query=select+distinct+*+where+{<PUT-THE-IRI-HERE>+<http%3A%2F%2Fwww.w3.org%2F2000%2F01%2Frdf-schema%23comment>+%3Fc+.%0D%0AFILTER(langMatches(lang(%3Fc)%2C+'EN'))%0D%0A%0D%0A}+LIMIT+100&format=json&timeout=30000"
	
	editor.on( 'nodechange', function( event ) {
		setState( event );
	});

	// Register a command so that it can be invoked by using tinyMCE.activeEditor.execCommand( 'WP_Entity_Link' );
	editor.addCommand( 'WP_Entity_Link', function() {
		
		var parentNode = editor.selection.getNode();
		
		var uriValue, labelValue;
		if ( parentNode.className == 'entityLink' && parentNode.getAttribute('data-uri').length > 0 ) {
			uriValue = parentNode.getAttribute('data-uri');
			labelValue = parentNode.getAttribute('data-label');
		} else {
			uriValue = '';
			labelValue = editor.selection.getContent();
		}

		var entityFormItems = [
			{id: "entitySearchInput", placeholder: 'Search for entities ..',label: 'Find Entities:', name: 'uriSearch', type: 'textbox', value: labelValue, onkeyup: updateAutoSuggestion, autofocus: true},
			{id: "entityUriLabel", type: 'label', text: uriValue},
			{id: "entityUri", name: 'uri', type: 'textbox', hidden: true, value: uriValue},
			{id: "entityAbstract", name: 'abstract', type: 'label', hidden: false, value: 'abstract' },
			{id: "entityLabel", name: 'label', type: 'textbox',  hidden: true, value: labelValue},
			{id: "entityCategory", name: 'category', hidden: true, type: 'textbox'}
		];

		window.entityWindow = editor.windowManager.open({
			title: 'Insert/edit entity',
			id: 'entityEditorWindow',
			body: entityFormItems,
			onSubmit: onSubmitForm
		});
		
		setTimeout(function() {
			jQuery('.mce-form.mce-first').append('<div id="autoSuggestionResult"></div>');
			var tmpEvt = {};
			tmpEvt.target = {};
			tmpEvt.target.value = labelValue;

			updateAutoSuggestion(tmpEvt);
			jQuery('#autoSuggestionResult').height(jQuery('#entityEditorWindow').height() - 250);
		}, 100);
		
		jQuery(window).resize(function() {
			jQuery('#autoSuggestionResult').height(jQuery('#entityEditorWindow').height() - 250);
		});
		
		function onSubmitForm(event) {
			
			if ( parentNode.className == 'entityLink' && event.data.uri === "" ) {
				
				editor.dom.remove(parentNode, true);

			} else if ( parentNode.className == 'entityLink' ) {
				
				parentNode.setAttribute('data-uri', event.data.uri);

			} else {

				var selected_text = editor.selection.getContent();
				var type = 'http://dbpedia.org/ontology/'+ event.data.category ;
                if (event.data.category=='http://www.w3.org/2002/07/owl#Thing'){
                        type = 'http://www.w3.org/2002/07/owl#Thing';
                }
                
                var return_text = '<span class="entityLink" property="http://www.w3.org/2004/02/skos/core#subject" data-createdby="user" typeof="' + type + '" resource="'+ event.data.uri +'" data-uri="'+ event.data.uri +'" data-category="'+ event.data.category +'" data-label="'+ event.data.label +'">' + selected_text + '</span>';

		        editor.execCommand('mceInsertContent', 0, return_text );

		    }
		}
		
		function updateAbstract(iri){
			jQuery.ajax({
				url: DBPEDIA.replace('PUT-THE-IRI-HERE', iri),
				crossDomain: true,
				beforeSubmit: function() {
					
				}
			})
			.done(function(data) {
				console.log(data);
				if (data.results.bindings.length > 0){
					jQuery('#entityAbstract').text( data.results.bindings[0].c.value );
				}else{
					jQuery('#entityAbstract').text( "" );
				}
				
			})
			.fail(function() {
				// silent
			})
			.always(function() {
				//
			});
			
		}

		function addToResultContainer(resultContainer, currentEntity){
			var catArray = currentEntity.categoryLabel.split(',');
			var catLabels = [];
			// TODO: make this replacements in suggestion service 
			for (var j = 0; j< catArray.length ; j++ ){
				var label = catArray[j].replace('http://schema.org/','').replace('http://xmlns.com/foaf/0.1/','');
				if (label != "http://www.w3.org/2002/07/owl#Thing"){
					if (catLabels.indexOf(label)==-1){
						catLabels.push( label );
					}
				}
			}
			var catLabelsStr = "";
			if (catLabels.length > 0) {
				catLabelsStr = " [" + catLabels.join(', ') +"]";
			}
			
			var imageString = '<span style="height: 18px; width:22px; background: rgba(0,0,0,.2); display: inline-block;"/>';
			if (currentEntity.thumbnailUri) {
				imageString = '<img style="height: auto; max-height:18px; width:22px;" src="' + currentEntity.thumbnailUri + '"/>';
			}
			var elem = jQuery('<a class="entitySuggestion" data-category="'+ currentEntity.categoryIri.replace('"http://www.w3.org/2002/07/owl#','') +'" title="'+ currentEntity.iri +'"><span style="width:30px; display:inline-block; ">' + imageString + '</span>'+ currentEntity.label + '<span style="font-size:10px; height: 16px; margin-left: 30px; display: block; overflow: hidden;">' + catLabelsStr + '</span></a>');
				elem.click( function() {
					jQuery('.entitySuggestion').removeClass('active');
					jQuery(this).addClass('active');
					jQuery('input#entityUri').val( jQuery(this).attr('title') );
					jQuery('#entityUriLabel').text( jQuery(this).attr('title') );
					
					jQuery('input#entityLabel').val( currentEntity.label );
					jQuery('input#entityCategory').val( jQuery(this).attr('data-category') );
					
					// fill into lastElements array (via tmp Array)
					// (1) put it into the new array  
					// (2) fill with other element entries, skipping the identical one (if existing) 					
					var tmp = [];
					tmp.push(currentEntity);
					for (var k = 0 ; k < Math.min(lastEntities.length, maxLastEntities) ; k++ ){
						if (currentEntity.iri != lastEntities[k].iri){
							tmp.push(lastEntities[k]);
						}
					}
					lastEntities = tmp;		
				});		
				elem.mouseover(  function(){ updateAbstract( currentEntity.iri ); })
				elem.appendTo(resultContainer);				
		}

		function updateAutoSuggestion(evt) {
			
			var resultContainer = jQuery('.mce-form.mce-first').children('#autoSuggestionResult');
			if (resultContainer.length == 0) {
				jQuery('.mce-form.mce-first').append('<div id="autoSuggestionResult"></div>');
			}
			jQuery.ajax({
				url: "http://apps.yovisto.com/refer-rest/services/suggest/x?input="+ evt.target.value.trim().replace("&","%26").split('.').join(''),
				crossDomain: true,
				beforeSubmit: function() {
					jQuery('input#entityUri').val('');
					jQuery('label#entityUriLabel').text('');
				}
			})
			.done(function(data) {
				resultContainer.empty();
				
				if (lastEntities.length > 0){
					var rsElem = jQuery('<h3 style="padding:5px;">Recently Selected:</h3>');
					rsElem.appendTo(resultContainer);
					
					var lastEntitiesContainer = jQuery('<div class="referLastEntitiesContainer"></div>');
					lastEntitiesContainer.appendTo(resultContainer);
				}
				
				// fill recently selected entities
				for (var i=0; i < lastEntities.length; i++){
					var currentEntity = lastEntities[i];
					addToResultContainer(lastEntitiesContainer, currentEntity);					
				}	
				// add separator
				if (lastEntities.length > 0){
					var elem = jQuery('<div style="padding:10px; clear: both;" />');
					elem.appendTo(resultContainer);
				}
				
				var sgElem = jQuery('<h3 style="padding:5px;">Suggestions:</h3>');
				sgElem.appendTo(resultContainer);
				
				// prepare the 4 columns				
				var columnPerson = jQuery('<div style="float:left; width:24.3%; margin-right: 1%;"><div style="padding: 5px;">Person</div></div>');
				var columnPlace = jQuery('<div style="float:left; width:24.3%; margin-right: 1%;"><div style="padding: 5px;">Place</div></div>');
				var columnThing = jQuery('<div style="float:left; width:24.2%; margin-right: 1%;"><div style="padding: 5px;">Thing</div></div>');
				var columnEvent = jQuery('<div style="float:left; width:24.2%; margin-right: 0%;"><div style="padding: 5px;">Event</div></div>');
				columnPerson.appendTo(resultContainer);
				columnPlace.appendTo(resultContainer);
				columnThing.appendTo(resultContainer);
				columnEvent.appendTo(resultContainer);
				
				// now add regular suggestions
				for (var i=0; i <  data.entities.length; i++) {
					var currentEntity = data.entities[i];
					if (currentEntity.categoryIri.replace('"http://www.w3.org/2002/07/owl#','')=='Person'){
						addToResultContainer(columnPerson, currentEntity);
					} else if (currentEntity.categoryIri.replace('"http://www.w3.org/2002/07/owl#','')=='Event'){
						addToResultContainer(columnEvent, currentEntity);
					} else if (currentEntity.categoryIri.replace('"http://www.w3.org/2002/07/owl#','')=='Place'){
						addToResultContainer(columnPlace, currentEntity);
					} else{					
						addToResultContainer(columnThing, currentEntity);
					}				
				}
			})
			.fail(function() {
				// silent
			})
			.always(function() {
				//
			});

		}

	});
	
	function getNextNode(node) {
	    if (node.firstChild)
	        return node.firstChild;
	    while (node) {
	        if (node.nextSibling)
	            return node.nextSibling;
	        node = node.parentNode;
	    }
	}

	function getNodesInRange(range) {
	    var start = range.startContainer;
	    var end = range.endContainer;
	    var commonAncestor = range.commonAncestorContainer;
	    var nodes = [];
	    var node;

	    // walk parent nodes from start to common ancestor
	    for (node = start.parentNode; node; node = node.parentNode) {
	        nodes.push(node);
	        if (node == commonAncestor)
	            break;
	    }
	    nodes.reverse();

	    // walk children and siblings from start until end is found
	    for (node = start; node; node = getNextNode(node)) {
	        nodes.push(node);
	        if (node == end)
	            break;
	    }

	    return nodes;
	}	

	editor.addCommand( 'WP_Entity_Unlink', function() {
		
		var parentNode = editor.selection.getNode();

		if ( parentNode.className == 'entityLink' ) {
			editor.dom.remove(parentNode, true);
		} else {
			var nodes = getNodesInRange(editor.selection.getRng());
			for ( var i=0; i < nodes.length; i++ ) {
				if ( nodes[i].className == 'entityLink' ) {
					editor.dom.remove(nodes[i], true);
				}
			}
		}

	});

	
	function setState( event ) {
		
		var node = event.element;
		
		var entityInSelection = false,
			nodes = getNodesInRange(editor.selection.getRng());

		for ( var i=0; i < nodes.length; i++ ) {
			if ( nodes[i].className == 'entityLink' ) {
				entityInSelection = true;
			}
		}
		
		
		if ( !editor.selection.isCollapsed() && node.className != 'entityLink' && !entityInSelection && !event.selectionChange) {			
			
			// clean up
			closeUnlinkMenu();

		} else if ( editor.selection.isCollapsed() && node.className != 'entityLink' && !entityInSelection ) {
			// clean up
			closeUnlinkMenu();

		} else if ( node.className == 'entityLink' && !node.name && !event.selectionChange ) {
			var textPosition = getSelectedTextPosition();
			openUnlinkMenu(textPosition);
		} else if ( entityInSelection && !node.name && !event.selectionChange ) {
			// Entity somewhere in selected text (dealt with by toolbar button)
		} else if ( event.selectionChange ) {
			// clean up
			closeUnlinkMenu();
		}
	}


	function setButtonState( button, event ) {
		
		var node = event.element;
		
		var entityInSelection = false;
		var nodes = getNodesInRange(editor.selection.getRng());
		for ( var i=0; i < nodes.length; i++ ) {
			if ( nodes[i].className == 'entityLink' ) {
				entityInSelection = true;
			}
		}
		
		
		if (button.settings.cmd == 'WP_Entity_Unlink' && node.className != 'entityLink' && !entityInSelection) {
			button.disabled(true);
		} else {
			button.disabled( editor.selection.isCollapsed() && node.className != 'entityLink' && !entityInSelection );
		}
		button.active( button.settings.cmd == 'WP_Entity_Unlink' && (node.className == 'entityLink' || entityInSelection) && ! node.name );
	}

	function openUnlinkMenu(position) {
		// clean up
		closeUnlinkMenu();

		var docBody = jQuery(editor.getContainer()).parents('body');

		// add html container elements
		var unlinkMenuContainer = jQuery('<div class="referUnlinkMenuContainer"></div>'),
			unlinkToolbar = jQuery('<div role="group" class="mce-toolbar-grp mce-inline-toolbar-grp mce-container mce-panel mce-arrow-down mce-arrow-left" style="position: relative;"></div>'),
			unlinkBody = jQuery('<div class="mce-toolbar mce-container-body mce-stack-layout"></div>'),
			unlinkButtonGroupWrapper = jQuery('<div role="toolbar" class="mce-container mce-stack-layout-item mce-first mce-last mce-btn-group"></div>'),
			//unlinkButtonEdit = jQuery('<div tabindex="-1" class="mce-widget mce-btn" role="button" aria-label="Edit "><button tabindex="-1" type="button" role="presentation"><i class="mce-ico mce-i-dashicon dashicons-edit"></i></button></div>'),
			unlinkButtonRemove = jQuery('<div tabindex="-1" class="mce-widget mce-btn mce-last" role="button" aria-label="Remove"><button tabindex="-1" type="button" role="presentation"><i class="mce-ico mce-i-dashicon dashicons-no"></i></button></div>'),
			nodeUri = jQuery(editor.selection.getNode()).attr('resource'),
			nodeCategory = jQuery(editor.selection.getNode()).data('category'),
			unlinkButtonUri = jQuery('<div style="display: inline-block; white-space: nowrap; font-size: 12px; line-height: 12px; padding: 8px 12px 8px 4px;">'+ nodeUri +'</div>');

		unlinkButtonGroupWrapper.append(unlinkButtonRemove, unlinkButtonUri);
		unlinkBody.append(unlinkButtonGroupWrapper);
		unlinkToolbar.append(unlinkBody);
		unlinkMenuContainer.append(unlinkToolbar);

		// init unlink menu
		unlinkButtonRemove.click(function(evt) {
			editor.execCommand('WP_Entity_Unlink');
		});

		if ( nodeCategory ) {
			unlinkMenuContainer.attr('data-category', nodeCategory);
		}

		unlinkMenuContainer.css({
			top: position.top + 'px',
			left: position.left + 'px'
		}).show();
		
		docBody.append(unlinkMenuContainer);
	}

	function closeUnlinkMenu() {
		jQuery(editor.getContainer()).parents('body').find('.referUnlinkMenuContainer').remove();
	}

	function getSelectedTextPosition() {
		var tinymcePosition = jQuery(editor.getContainer()).position(),
			toolbarPosition = jQuery(editor.getContainer()).find(".mce-toolbar").first(),
			nodePosition = jQuery(editor.selection.getNode()).position(),
			textareaTop = 0,
			textareaLeft = 0;

			if (editor.selection.getRng().getClientRects().length > 0) {
			    textareaTop = parseInt(jQuery(editor.selection.getNode()).css("font-size")) * 0.4 + editor.selection.getRng().getClientRects()[0].top;
			    textareaLeft = editor.selection.getRng().getClientRects()[0].left;
			} else {
			    textareaTop = parseInt(jQuery(editor.selection.getNode()).css("font-size")) * 0.4 + nodePosition.top;
			    textareaLeft = nodePosition.left;
			}

			var position = jQuery(editor.getContainer()).offset();
			var caretPosition = {
			    top:  tinymcePosition.top + toolbarPosition.innerHeight() + textareaTop + position.top,
			    left: tinymcePosition.left + textareaLeft + position.left
			}

		return caretPosition;
	}

	editor.addButton( 'linkEntity', {
		icon: 'linkEntity',
		tooltip: 'Insert/edit entity',
		shortcut: 'Alt+Shift+A',
		cmd: 'WP_Entity_Link',

		onPostRender: function() {
			linkButton = this;

			editor.on( 'nodechange', function( event ) {
				setButtonState( linkButton, event );
			});
		}
	});

	editor.addButton( 'unlinkEntity', {
		icon: 'unlinkEntity',
		tooltip: 'Remove entity',
		cmd: 'WP_Entity_Unlink',
		onPostRender: function() {
			var unlinkButton = this;

			editor.on( 'nodechange', function( event ) {
				setButtonState(unlinkButton, event);
			});
		}
	});
});
