<?php
/*
Plugin Name: refer
Plugin Script: refer.php
Plugin URI: http://refer.cx
Description: Interactive Recommendation Plugin based on Linked Open Data and Semantic Web Technologies. 
Version: 0.9
Author: Joscha Jaeger
Author URI: http://filmicweb.org
Copyright: 2016 Joscha Jaeger
License: GPLv2 or later
License URI: http://www.gnu.org/licenses/gpl-2.0.html

=== RELEASE NOTES ===
2016-06-07 - v0.9 - Inline Editing
2015-04-29 - v0.8 - Beta Release
*/

/************************************
* Options
************************************/

$navigatorVisible = get_option('refer_visible');

/************************************
* Compatibility
************************************/

// set word limit to default value (6), if it doesn't exist
if ( strlen(get_option('refer_word_limit')) == 0 ) {
  update_option('refer_word_limit', '6');
}

// set annotation interface to default value, if it doesn't exist
if ( strlen(get_option('refer_annotation_interface')) == 0 ) {
  update_option('refer_annotation_interface', 'inline');
}

/************************************
* Includes
************************************/

if ($navigatorVisible) {
	include_once(dirname(__FILE__).'/interface/navigator-interface-functions.php');
}

include_once(dirname(__FILE__).'/adminpanel/navigator-adminpanel-functions.php');
include_once(dirname(__FILE__).'/editor/navigator-editor-functions.php');

/************************************
* Inject CSS & JS for refer Interface
************************************/

if ($navigatorVisible) {
	add_action('get_header', 'refer_add_scripts');
}

add_action( 'admin_enqueue_scripts', 'refer_add_editor_scripts' );

add_filter( 'mce_css', 'plugin_mce_css' );

add_filter('show_admin_bar', '__return_false');

/************************************
* Insert refer Interface just before </body>
************************************/

if ($navigatorVisible) {
	add_action('wp_footer', 'refer');
}

/************************************
* Add Buttons to Editor
************************************/

add_action('init', 'refer_add_editor_buttons');
add_action('save_post', 'refer_add_updateservice');
add_action( 'before_delete_post', 'refer_add_deleteservice');
add_action('trash_post','refer_add_deleteservice');

/************************************
* Add refer Admin Panel to "Settings"
************************************/

add_action('admin_menu', 'refer_add_adminpanel');

/************************************
* Plugin Activation Routine
************************************/

register_activation_hook(__FILE__, 'refer_exec_install');

function refer_exec_install() {
	$notices= get_option('refer_deferred_admin_notices', array());
	$notices[]= "Thank you for installing <b>refer</b>! Now, please go to the <a href='". admin_url( 'options-general.php?page=refer_options', 'http' ) ."'>Settings Page</a> and paste your API Key";
	update_option('refer_deferred_admin_notices', $notices);
	add_option('refer_initial_entity', 'http://dbpedia.org/resource/Germany');
	add_option('refer_visible', true);
	add_option('refer_theme', 'light');
  add_option('refer_annotation_interface', 'inline');
  // needs to be a string (will be parsed as int in JS)
  add_option('refer_word_limit', '6');
}

// Plugin Update Routine

/*
add_action('admin_init', 'refer_admin_init');
function refer_admin_init() {
  $current_version = 1;
  $version = get_option('refer_version');
  if ($version != $current_version) {
    update_option('refer_version', $current_version);
    $notices= get_option('refer_deferred_admin_notices', array());
    $notices[]= "refer: Upgraded version $version to $current_version.";
    update_option('refer_deferred_admin_notices', $notices);
  }
}
*/

add_action('admin_notices', 'refer_admin_notices');

function refer_admin_notices() {
  if ($notices = get_option('refer_deferred_admin_notices')) {
    foreach ($notices as $notice) {
      echo "<div class='update-nag'><p>$notice</p></div>";
    }
    delete_option('refer_deferred_admin_notices');
  }
}

/************************************
* Plugin Deactivation Routine
************************************/

register_deactivation_hook(__FILE__, 'refer_deactivation');

function refer_deactivation() {
	delete_option('refer_version'); 
	delete_option('refer_deferred_admin_notices'); 
	//delete_option('refer_api_key');
	delete_option('refer_initial_entity');
	delete_option('refer_visible');
	delete_option('refer_theme');
  delete_option('refer_word_limit');
  delete_option('refer_annotation_interface');
}

/************************************
* Extend Valid MCE Elements for RDFa
************************************/

function extend_valid_mce_elements( $init ) {
    $rdfa = '@[property|rel|about|resource|rev|typeof|content|datatype|title|class|data-category|data-category],span,div';
    $init['extended_valid_elements'] = $rdfa;
    return $init;
}
add_filter('tiny_mce_before_init', 'extend_valid_mce_elements');
?>